@@ -1,4 +1,5 @@
 #include <sys/stat.h>
+#include <sys/types.h>
 
 #include <err.h>
 #include <errno.h>
@@ -69,17 +70,17 @@ static const char *cachefile;
 #define pledge(p1,p2) 0
 #endif
 
-/* print `len' columns of characters. If string is shorter pad the rest
+/* format `len' columns of characters. If string is shorter pad the rest
  * with characters `pad`. */
-void
-printutf8pad(FILE *fp, const char *s, size_t len, int pad)
+int
+utf8pad(char *buf, size_t bufsiz, const char *s, size_t len, int pad)
 {
 	wchar_t w;
-	size_t col = 0, i, slen;
+	size_t col = 0, i, slen, siz = 0;
 	int rl, wc;
 
 	if (!len)
-		return;
+		return -1;
 
 	slen = strlen(s);
 	for (i = 0; i < slen && col < len + 1; i += rl) {
@@ -87,15 +88,28 @@ printutf8pad(FILE *fp, const char *s, size_t len, int pad)
 			break;
 		if ((wc = wcwidth(w)) == -1)
 			wc = 1;
-		col += (size_t)wc;
+		col += wc;
 		if (col >= len && s[i + rl]) {
-			fputs("\xe2\x80\xa6", fp);
-			break;
+			if (siz + 4 >= bufsiz)
+				return -1;
+			memcpy(&buf[siz], "\xe2\x80\xa6", 4);
+			return 0;
 		}
-		fwrite(&s[i], 1, rl, fp);
+		if (siz + rl + 1 >= bufsiz)
+			return -1;
+		memcpy(&buf[siz], &s[i], rl);
+		siz += rl;
+		buf[siz] = '\0';
 	}
-	for (; col < len; col++)
-		putc(pad, fp);
+
+	len -= col;
+	if (siz + len + 1 >= bufsiz)
+		return -1;
+	memset(&buf[siz], pad, len);
+	siz += len;
+	buf[siz] = '\0';
+
+	return 0;
 }
 
 void
@@ -254,7 +268,7 @@ efopen(const char *name, const char *flags)
 	FILE *fp;
 
 	if (!(fp = fopen(name, flags)))
-		err(1, "fopen");
+		err(1, "fopen: '%s'", name);
 
 	return fp;
 }
@@ -277,35 +291,6 @@ xmlencode(FILE *fp, const char *s, size_t len)
 	}
 }
 
-void
-trim(char *buf, size_t bufsiz, const char *src)
-{
-	size_t d = 0, i, len, s;
-
-	len = strlen(src);
-	for (s = 0; s < len && d < bufsiz - 1; s++) {
-		switch (src[s]) {
-		case '\t':
-			if (d + 8 >= bufsiz - 1)
-				goto end;
-			for (i = 0; i < 8; i++)
-				buf[d++] = ' ';
-			break;
-		case '\r': /* ignore CR */
-		case '|': /* ignore separators here */
-			break;
-		case '\n':
-			buf[d++] = ' ';
-			break;
-		default:
-			buf[d++] = src[s];
-			break;
-		}
-	}
-end:
-	buf[d] = '\0';
-}
-
 /* Escape characters in text in geomyidae .gph format, with newlines */
 void
 gphtextnl(FILE *fp, const char *s, size_t len)
@@ -316,7 +301,7 @@ gphtextnl(FILE *fp, const char *s, size_t len)
 		if (s[i] == '\n')
 			n = 0;
 
-		/* escape 't' at the start of a line */
+		/* escape with 't' at the start of a line */
 		if (!n && (s[i] == 't' || s[i] == '[')) {
 			fputc('t', fp);
 			n = 1;
@@ -340,10 +325,15 @@ gphtext(FILE *fp, const char *s, size_t len)
 
 	for (i = 0; *s && i < len; i++) {
 		switch (s[i]) {
-		case '\r':
-		case '\n': break;
-		case '\t': fputs("        ", fp); break;
-		default: fputc(s[i], fp);
+		case '\r': /* ignore CR */
+		case '\n': /* ignore LF */
+			break;
+		case '\t':
+			fputs("        ", fp);
+			break;
+		default:
+			fputc(s[i], fp);
+			break;
 		}
 	}
 }
@@ -356,16 +346,15 @@ gphlink(FILE *fp, const char *s, size_t len)
 
 	for (i = 0; *s && i < len; i++) {
 		switch (s[i]) {
-		case '\n':
-			/* in this context replace newline with space */
-			fputc(' ', fp);
-			break;
 		case '\r': /* ignore CR */
-		case '|': /* ignore separators here */
+		case '\n': /* ignore LF */
 			break;
 		case '\t':
 			fputs("        ", fp);
 			break;
+		case '|': /* escape separators */
+			fputs("\\|", fp);
+			break;
 		default:
 			fputc(s[i], fp);
 			break;
@@ -575,12 +564,12 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 		if (strcmp(delta->old_file.path, delta->new_file.path)) {
 			snprintf(filename, sizeof(filename), "%s -> %s",
 				delta->old_file.path, delta->new_file.path);
-			trim(buf, sizeof(buf), filename);
+			utf8pad(buf, sizeof(buf), filename, 35, ' ');
 		} else {
-			trim(buf, sizeof(buf), delta->old_file.path);
+			utf8pad(buf, sizeof(buf), delta->old_file.path, 35, ' ');
 		}
 		fputs("  ", fp);
-		printutf8pad(fp, buf, 35, ' ');
+		gphtext(fp, buf, strlen(buf));
 
 		add = ci->deltas[i]->addcount;
 		del = ci->deltas[i]->delcount;
@@ -655,16 +644,14 @@ writelogline(FILE *fp, struct commitinfo *ci)
 	fputs("[1|", fp);
 	if (ci->author)
 		printtimeshort(fp, &(ci->author->when));
+	else
+		fputs("                ", fp);
 	fputs("  ", fp);
-	if (ci->summary) {
-		trim(buf, sizeof(buf), ci->summary);
-		printutf8pad(fp, buf, 50, ' ');
-	}
+	utf8pad(buf, sizeof(buf), ci->summary ? ci->summary : "", 50, ' ');
+	gphlink(fp, buf, strlen(buf));
 	fputs("  ", fp);
-	if (ci->author) {
-		trim(buf, sizeof(buf), ci->author->name);
-		printutf8pad(fp, buf, 25, ' ');
-	}
+	utf8pad(buf, sizeof(buf), ci->author ? ci->author->name : "", 25, ' ');
+	gphlink(fp, buf, strlen(buf));
 	fprintf(fp, "|%s/commit/%s.gph", relpath, ci->oid);
 	fputs("|server|port]\n", fp);
 }
@@ -926,8 +913,8 @@ writefilestree(FILE *fp, git_tree *tree, const char *path)
 			fputs("[1|", fp);
 			fputs(filemode(git_tree_entry_filemode(entry)), fp);
 			fputs("  ", fp);
-			trim(buf, sizeof(buf), entrypath);
-			printutf8pad(fp, buf, 50, ' ');
+			utf8pad(buf, sizeof(buf), entrypath, 50, ' ');
+			gphlink(fp, buf, strlen(buf));
 			fputs("  ", fp);
 			if (lc > 0)
 				fprintf(fp, "%7dL", lc);
@@ -938,8 +925,8 @@ writefilestree(FILE *fp, git_tree *tree, const char *path)
 			git_object_free(obj);
 		} else if (!git_submodule_lookup(&module, repo, entryname)) {
 			fputs("[1|m---------  ", fp);
-			trim(buf, sizeof(buf), entrypath);
-			printutf8pad(fp, buf, 50, ' ');
+			utf8pad(buf, sizeof(buf), entrypath, 50, ' ');
+			gphlink(fp, buf, strlen(buf));
 			fprintf(fp, "|%s/file/.gitmodules.gph|server|port]\n", relpath);
 			/* NOTE: linecount omitted */
 			git_submodule_free(module);
@@ -1042,21 +1029,23 @@ writerefs(FILE *fp)
 				fprintf(fp, "%s\n", titles[j]);
 				fprintf(fp, "  %-20.20s", "Name");
 				fprintf(fp, "  %-16.16s", "Last commit date");
-				fprintf(fp, "  %-25.25s\n", "Author");
+				fprintf(fp, "  %s\n", "Author");
 			}
 
 			name = git_reference_shorthand(r);
 
 			fputs("  ", fp);
-			trim(buf, sizeof(buf), name);
-			printutf8pad(fp, buf, 20, ' ');
+			utf8pad(buf, sizeof(buf), name, 20, ' ');
+			gphlink(fp, buf, strlen(buf));
 			fputs("  ", fp);
 			if (ci->author)
 				printtimeshort(fp, &(ci->author->when));
+			else
+				fputs("                ", fp);
 			fputs("  ", fp);
 			if (ci->author) {
-				trim(buf, sizeof(buf), ci->author->name);
-				printutf8pad(fp, buf, 25, ' ');
+				utf8pad(buf, sizeof(buf), ci->author->name, 25, ' ');
+				gphlink(fp, buf, strlen(buf));
 			}
 			fputs("\n", fp);
 
@@ -1095,6 +1084,7 @@ main(int argc, char *argv[])
 	git_object *obj = NULL;
 	const git_oid *head = NULL;
 	const git_error *e = NULL;
+	mode_t mask;
 	FILE *fp, *fpread;
 	char path[PATH_MAX], repodirabs[PATH_MAX + 1], *p;
 	char tmppath[64] = "cache.XXXXXXXXXXXX", buf[BUFSIZ];
@@ -1199,12 +1189,12 @@ main(int argc, char *argv[])
 
 	/* log for HEAD */
 	fp = efopen("log.gph", "w");
-	mkdir("commit", 0755);
+	mkdir("commit", S_IRWXU | S_IRWXG | S_IRWXO);
 	writeheader(fp, "Log");
 
 	fprintf(fp, "%-16.16s  ", "Date");
 	fprintf(fp, "%-50.50s  ", "Commit message");
-	fprintf(fp, "%-25.25s\n", "Author");
+	fprintf(fp, "%s\n", "Author");
 
 	if (cachefile) {
 		/* read from cache file (does not need to exist) */
@@ -1219,7 +1209,7 @@ main(int argc, char *argv[])
 		if ((fd = mkstemp(tmppath)) == -1)
 			err(1, "mkstemp");
 		if (!(wcachefp = fdopen(fd, "w")))
-			err(1, "fdopen");
+			err(1, "fdopen: '%s'", tmppath);
 		/* write last commit id (HEAD) */
 		git_oid_tostr(buf, sizeof(buf), head);
 		fprintf(wcachefp, "%s\n", buf);
@@ -1268,8 +1258,14 @@ main(int argc, char *argv[])
 	fclose(fp);
 
 	/* rename new cache file on success */
-	if (cachefile && rename(tmppath, cachefile))
-		err(1, "rename: '%s' to '%s'", tmppath, cachefile);
+	if (cachefile) {
+		if (rename(tmppath, cachefile))
+			err(1, "rename: '%s' to '%s'", tmppath, cachefile);
+		umask((mask = umask(0)));
+		if (chmod(cachefile,
+		    (S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH) & ~mask))
+			err(1, "chmod: '%s'", cachefile);
+	}
 
 	/* cleanup */
 	git_repository_free(repo);
